#include <errno.h>
#include <stdio.h>

#include <linux/nl80211.h>
#include <netlink/netlink.h>
#include <netlink/genl/ctrl.h>
#include <netlink/genl/family.h>
#include <netlink/genl/genl.h>

#include "wifi.h"

static struct nla_policy stats_policy[NL80211_STA_INFO_MAX + 1] = {
	[NL80211_STA_INFO_INACTIVE_TIME] = { .type = NLA_U32 },
	[NL80211_STA_INFO_RX_BYTES] = { .type = NLA_U32 },
	[NL80211_STA_INFO_TX_BYTES] = { .type = NLA_U32 },
	[NL80211_STA_INFO_RX_PACKETS] = { .type = NLA_U32 },
	[NL80211_STA_INFO_TX_PACKETS] = { .type = NLA_U32 },
	[NL80211_STA_INFO_SIGNAL] = { .type = NLA_U8 },
	[NL80211_STA_INFO_TX_BITRATE] = { .type = NLA_NESTED },
	[NL80211_STA_INFO_LLID] = { .type = NLA_U16 },
	[NL80211_STA_INFO_PLID] = { .type = NLA_U16 },
	[NL80211_STA_INFO_PLINK_STATE] = { .type = NLA_U8 },
};

static struct nla_policy rate_policy[NL80211_RATE_INFO_MAX + 1] = {
	[NL80211_RATE_INFO_BITRATE] = { .type = NLA_U16 },
	[NL80211_RATE_INFO_MCS] = { .type = NLA_U8 },
	[NL80211_RATE_INFO_40_MHZ_WIDTH] = { .type = NLA_FLAG },
	[NL80211_RATE_INFO_SHORT_GI] = { .type = NLA_FLAG },
};


int init_nl80211(Netlink *nl, Wifi *w)
{
	nl->socket = nl_socket_alloc();
	if (!nl->socket) {
		fprintf(stderr, "failed to alloc netlink socket\n");
		return -ENOMEM;
	}

	nl_socket_set_buffer_size(nl->socket, 8192, 8192);

	if (genl_connect(nl->socket)) {
		fprintf(stderr, "failed to connect to netlink socket\n");
		nl_close(nl->socket);
		nl_socket_free(nl->socket);
		return -ENOLINK;
	}

	nl->id = genl_ctrl_resolve(nl->socket, "nl80211");
	if (nl->id < 0) {
		fprintf(stderr, "nl80211 interface not found\n");
		nl_close(nl->socket);
		nl_socket_free(nl->socket);
		return -ENOENT;
	}

	nl->cb1 = nl_cb_alloc(NL_CB_DEFAULT);
	nl->cb2 = nl_cb_alloc(NL_CB_DEFAULT);

	if ((!nl->cb1) || (!nl->cb2)) {
		fprintf(stderr, "failed to alloc netlink callbacks\n");
		nl_close(nl->socket);
		nl_socket_free(nl->socket);
		return -ENOMEM;
	}

	nl_cb_set(nl->cb1, NL_CB_VALID , NL_CB_CUSTOM, get_wifi_name_cb, w);
	nl_cb_set(nl->cb1, NL_CB_FINISH, NL_CB_CUSTOM, nl80211_finish_handler,
		      &(nl->result1));

	nl_cb_set(nl->cb2, NL_CB_VALID , NL_CB_CUSTOM, get_wifi_info_cb, w);
	nl_cb_set(nl->cb2, NL_CB_FINISH, NL_CB_CUSTOM, nl80211_finish_handler,
		      &(nl->result2));

	return nl->id;
}

int nl80211_finish_handler(struct nl_msg *msg, void *arg)
{
	(void)msg;

	int *ret = arg;
	*ret = 0;
	return NL_SKIP;
}

int get_wifi_name_cb(struct nl_msg *msg, void *arg)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));

	nl_msg_dump(msg, stdout);

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		      genlmsg_attrlen(gnlh, 0), NULL);

	if (tb[NL80211_ATTR_IFNAME])
		strcpy(((Wifi*)arg)->ifname, nla_get_string(tb[NL80211_ATTR_IFNAME]));

	if (tb[NL80211_ATTR_IFINDEX])
		((Wifi*)arg)->ifindex = nla_get_u32(tb[NL80211_ATTR_IFINDEX]);

	return NL_SKIP;
}

int get_wifi_info_cb(struct nl_msg *msg, void *arg)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct nlattr *sinfo[NL80211_STA_INFO_MAX + 1];
	struct nlattr *rinfo[NL80211_RATE_INFO_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));

	nl_msg_dump(msg, stdout);

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		      genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_STA_INFO]) {
		fprintf(stderr, "sta stats missing\n");
		return NL_SKIP;
	}

	if (nla_parse_nested(sinfo, NL80211_STA_INFO_MAX,
		                 tb[NL80211_ATTR_STA_INFO], stats_policy))
	{
		fprintf(stderr, "failed to parse nested attrs\n");
		return NL_SKIP;
	}

	if (sinfo[NL80211_STA_INFO_SIGNAL])
		((Wifi*)arg)->signal = 100+(int8_t)nla_get_u8(sinfo[NL80211_STA_INFO_SIGNAL]);

	if (sinfo[NL80211_STA_INFO_TX_BITRATE]) {
		if (nla_parse_nested(rinfo, NL80211_RATE_INFO_MAX,
			                 sinfo[NL80211_STA_INFO_TX_BITRATE], rate_policy))
		{
			fprintf(stderr, "failed to parse nested rate attrs\n");
		} else {
			if (rinfo[NL80211_RATE_INFO_BITRATE])
				((Wifi*)arg)->txrate = nla_get_u16(rinfo[NL80211_RATE_INFO_BITRATE]);
		}
	}

	return NL_SKIP;
}

int get_wifi_status(Netlink *nl, Wifi *w)
{
	nl->result1 = 1;
	nl->result2 = 2;

	struct nl_msg *msg1 = nlmsg_alloc();
	if (!msg1) {
		fprintf(stderr, "failed to alloc netlink msg1\n");
		return -2;
	}

	genlmsg_put(msg1, NL_AUTO_PORT, NL_AUTO_SEQ, nl->id, 0, NLM_F_DUMP,
		        NL80211_CMD_GET_INTERFACE, 0);

	nl_send_auto(nl->socket, msg1);

	while (nl->result1 > 0)
		nl_recvmsgs(nl->socket, nl->cb1);

	nlmsg_free(msg1);

	if (w->ifindex < 0)
		return -1;

	struct nl_msg *msg2 = nlmsg_alloc();
	if (!msg2) {
		fprintf(stderr, "failed to alloc netlink msg2\n");
		return -2;
	}

	genlmsg_put(msg2, NL_AUTO_PORT, NL_AUTO_SEQ, nl->id, 0, NLM_F_DUMP,
		        NL80211_CMD_GET_STATION, 0);

	nla_put_u32(msg2, NL80211_ATTR_IFINDEX, w->ifindex);
	nl_send_auto(nl->socket, msg2);

	while (nl->result2 > 0)
		nl_recvmsgs(nl->socket, nl->cb2);

	nlmsg_free(msg2);

	return 0;
}
